/**
 * Introduction
 *
 * @title ROVLib - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file alarm.h
 *
 * @type subsystem
 *
 * @description
 *
 * The alarm subsystem is used to maintain 'alarms' that are designed to go
 * off at particular times, based on the OS_MonotonicTime clock.
 *
 * For example, it can be used as a periodic poll, say to flash a cursor, or
 * to maintain a socket, or as a once-off poll such as a message box that
 * appears on the screen.
 *
 * Each alarm that is set has three parameters - the time at which it is due
 * to go off, the function to call, and a handle that the function can use to
 * identify the alarm itself.
 *
 * Once the time has been passed for the alarm to go off has been reached,
 * the alarm handler function is called. Note that this is the minimum time 
 * that the function should be called, and can occur after the alarm was due.
 */

#ifndef __rov_alarm
#define __rov_alarm


#ifdef __cplusplus
extern "C"
{
#endif

/**
 * Definition of an alarm handler.
 * @param time the time that the alarm was due to go off.
 * @param handle the alarm handle value.
 * @see alarm_set()
 */
typedef void (*alarm_proc)(int time,
			   int handle);

/**
 * Initialises the alarm subsystem. Do this before calling main_poll
 */
void alarm_init(void);

/**
 * Set an alarm.
 * @param time the time the alarm is due to go off.
 * @param proc the alarm handler.
 * @param handle the alarm handle value.
 * @see alarm_proc()
 * @see alarm_cancel()
 */
void alarm_set(int time,
	       alarm_proc proc,
	       int handle);

/**
 * Cancels an alarm.
 * @param handle the alarm handle value. This will cancel all alarms with this
 * handle value.
 * @see alarm_set()
 */
void alarm_cancel(int handle);


#ifdef __cplusplus
}
#endif

#endif
