/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file flex.c
 */

#include <stdio.h>
#include <string.h>

#include "kernel.h"

#include "main.h"
#include "dynamic.h"
#include "omemcpy.h"


static _kernel_swi_regs regs;

typedef struct flex_block {
  struct flex_block * next;
  char * * back_ptr;
  int size;
  char data[1];
} flex_block;

typedef void (* flex_reg)(int);

#define FLEX_PRERESIZE  (-1)
#define FLEX_POSTRESIZE (-2)
#define FLEX_ALLOC      (0)
#define FLEX_FREE       (1)
#define FLEX_RESIZE     (2)
#define FLEX_CLEAR      (3)
#define FLEX_SWAP       (4)
#define FLEX_MOVE       (5)
#define FLEX_USER       (100)

#define WSE 0x600ec

static flex_block * first = 0;

static void flex_nullreg(int a)
{
}

int flex_alloc(char * * b, int size);

static int allocated = 0, init_alloc = 0, memory_free = 0, memory_used = 0;
static flex_block * last = 0;
static char * rmalloc_block;
static flex_reg _reg = flex_nullreg;
static int dynamic_area;
static int dynamic_limit = -1;

static int rmalloc_extend(int);
static void * rmalloc_init(void);

void rmalloc_claimhandlers(int (*e)(int), void *(*i)(void));

void flex_limitdyn(int size)
{
  dynamic_limit = size;
}

void flex_register(flex_reg r)
{
  if(r) _reg = r; else _reg = flex_nullreg;
}

void flex_callregister(int i)
{
  _reg(i);
}

void flex_init(void)
{
  regs.r[0] = -1;
  regs.r[1] = -1;
  _kernel_swi(WSE, & regs, & regs);
  init_alloc = regs.r[0];
  rmalloc_claimhandlers(rmalloc_extend, rmalloc_init);
  rmalloc_init();
}

void flex_cleardyn(void)
{
  dynamicarea_destroy(dynamic_area);
  dynamic_area = 0;
}

void flex_initdyn(void)
{
  static char flexname[32];

  strcpy(flexname, "ROVflex-");
  strncpy(flexname + 8, TASKNAME, 23);
  dynamic_area = dynamicarea_create(flexname, 0, dynamic_limit);
  if(dynamic_area == 0) {
    flex_init();
    rmalloc_newinit();
  } else {
    init_alloc = (int)dynamicarea_address(dynamic_area) - 0x8000;
    atexit(flex_cleardyn);
  }
}

int flex_max(void)
{
  regs.r[0] = -1;
  regs.r[1] = -1;
  _kernel_swi(WSE, & regs, & regs);
  return regs.r[2] + memory_free - 48;
}

static void * rmalloc_init(void)
{
  if(rmalloc_block) return (void *)rmalloc_block;
  if(flex_alloc(& rmalloc_block, 0) == 0) return 0;
  return (void *)rmalloc_block;
}

static int allocate_more(int size)
{
  regs.r[0] = -1;
  regs.r[1] = -1;
  _kernel_swi(WSE, & regs, & regs);
  if(regs.r[2] < size + 24)
    return 0;
  _reg(FLEX_PRERESIZE);
  if(dynamic_area) {
    if(dynamicarea_extend(dynamic_area, allocated + size + 24) == 0) return 0;
  } else {
    regs.r[0] = init_alloc + allocated + size + 24;
    regs.r[1] = -1;
    _kernel_swi(WSE, & regs, & regs);
    if(regs.r[0] < init_alloc + allocated + size + 24) {
      regs.r[0] = init_alloc + allocated;
      regs.r[1] = -1;
      _kernel_swi(WSE, & regs, & regs);
      return 0;
    }
  }
  _reg(FLEX_POSTRESIZE);
  if(dynamic_area) {
    allocated = dynamicarea_size(dynamic_area);
  } else {
    allocated = regs.r[0] - init_alloc;
  }
  memory_free = allocated - memory_used;
  return 1;
}

int flex_alloc(char * * b, int size)
{
  flex_block * next;

  size = (size + 3) & 0xfffffffc;

  if(memory_free < size + 28)
    if(allocate_more(size - memory_free + 28) == 0)
      return 0;

  if(first == 0)
    first = last = (flex_block *)(init_alloc + 0x8000);

  memory_free -= size + 12;
  memory_used += size + 12;
  * b = & last -> data[0];
  last -> back_ptr = b;
  last -> size = size;
  next = (flex_block *)((int)last + size + 12);
  last -> next = next;
  last = next;
  last -> next = 0;
  last -> size = 0;
  last -> back_ptr = 0;
  _reg(FLEX_ALLOC);
  return 1;
}

int flex_allocmax(char * * b)
{
  flex_block * next;
  int size;

  size = flex_max();

  if(memory_free < size + 12)
    if(allocate_more(size - memory_free + 12) == 0) {
      size = flex_max();
      return flex_alloc(b, size);
    }

  if(first == 0)
    first = last = (flex_block *)(init_alloc + 0x8000);

  memory_free -= size + 12;
  memory_used += size + 12;
  * b = & last -> data[0];
  last -> back_ptr = b;
  last -> size = size;
  next = (flex_block *)((int)last + size + 12);
  last -> next = next;
  last = next;
  last -> next = 0;
  last -> size = 0;
  last -> back_ptr = 0;
  _reg(FLEX_ALLOC);
  return 1;
}

static void reshuffle_used(void)
{
  _reg(FLEX_PRERESIZE);
  if(dynamic_area) {
    dynamicarea_extend(dynamic_area, memory_used + 24);
  } else {
    regs.r[0] = init_alloc + memory_used + 24;
    regs.r[1] = -1;
    _kernel_swi(WSE, & regs, & regs);
  }
  _reg(FLEX_POSTRESIZE);
  if(dynamic_area) {
    allocated = dynamicarea_size(dynamic_area);
  } else {
    allocated = regs.r[0] - init_alloc;
  }
  memory_free = allocated - memory_used;
  if(allocated == 0)
    first = last = 0;
}

static void change_pointers(flex_block * fl, int difference)
{
  while(fl != last) {
    fl -> next = (flex_block *)((int)fl -> next + difference);
    fl = fl -> next;
    if(fl != last)
      * fl -> back_ptr = (char *)((int)(* fl -> back_ptr) + difference);
  }
}

static void change_pointersdown(flex_block * fl, int difference)
{
  while(fl != last) {
    fl -> next = (flex_block *)((int)fl -> next - difference);
    * fl -> back_ptr = (char *)((int)(* fl -> back_ptr) - difference);
    fl = fl -> next;
  }
}

void flex_free(char * * b)
{
  flex_block * thisone;
  int difference;

  if(b == 0) return;

  thisone = (flex_block *)((int)(* b) - 12);
  if(thisone -> back_ptr != b)
    return; /* Whoops ! */
  *b = 0;
  difference = thisone -> size + 12;
  if(thisone -> next != last) {
    omemcpy(thisone, thisone -> next, (int)last - (int)thisone -> next);
    /*MemoryCopy(thisone -> next, thisone, (int)last - (int)thisone -> next);*/
    last = (flex_block *)((int)last - difference);
    change_pointersdown(thisone, difference);
  } else {
    last = thisone;
  }

  last -> back_ptr = 0;
  last -> next = 0;
  last -> size = 0;
  memory_used -= difference;
  memory_free += difference;
  reshuffle_used();
  _reg(FLEX_FREE);
}

int flex_extendby(char * * b, int size)
{
  flex_block * thisone = (flex_block *)((int)(* b) - 12);

  size = (size + 3) & 0xfffffffc;

  if(thisone -> back_ptr != b)
    return 0;
  if(size < 0)
    goto decrease_size;

  if(memory_free < size + 12)
    if(allocate_more(size - memory_free + 12) == 0)
      return 0;
  thisone -> size += size;
  memory_used += size;
  memory_free -= size;
  if(thisone -> next != last) {
    omemcpy((void *)((int)thisone -> next + size), thisone -> next, (int)last - (int)thisone -> next);
    /*MemoryCopy(thisone -> next, (void *)((int)thisone -> next + size), (int)last - (int)thisone -> next);*/
    last = (flex_block *)((int)last + size);
    change_pointers(thisone, size);
  } else {
    thisone -> next = last = (flex_block *)((int)last + size);
  }

  last -> back_ptr = 0;
  last -> next = 0;
  last -> size = 0;
  _reg(FLEX_RESIZE);
  return 1;

  decrease_size :
    if(thisone -> size + size <= 0) {
      flex_free(b);
      return 1;
    }
  thisone -> size += size;
  memory_used += size;
  memory_free -= size;
  if(thisone -> next != last) {
    omemcpy((void *)((int)thisone -> next + size), thisone -> next, (int)last - (int)thisone -> next);
    /*MemoryCopy(thisone -> next, (void *)((int)thisone -> next + size), (int)last - (int)thisone -> next);*/
    last = (flex_block *)((int)last + size);
    change_pointers(thisone, size);
  } else {
    thisone -> next = last = (flex_block *)((int)last + size);
  }

  last -> back_ptr = 0;
  last -> next = 0;
  last -> size = 0;
  reshuffle_used();
  _reg(FLEX_RESIZE);
  return 1;
}

int flex_size(char * * a)
{
  flex_block * af = (flex_block *)((int)(* a) - 12);

  if(af -> back_ptr != a)
    return 0;
  return af -> size;
}

int flex_extend(char * * a, int to)
{
  return flex_extendby(a, to - flex_size(a));
}

void flex_swap(char * * a, char * * b)
{
  flex_block * af = (flex_block *)((int)(* a) - 12), * bf = (flex_block *)((int)(* b) - 12);
  char * * c = a;

  if((af -> back_ptr != a) || (bf -> back_ptr != b))
    return;

  af -> back_ptr = b;
  bf -> back_ptr = c;
  * c = & bf -> data[0];
  * b = & af -> data[0];
  _reg(FLEX_SWAP);
}

void flex_move(char * * a, char * * b)
{
  flex_block * af = (flex_block *)((int)(* a) - 12);
  if(*b) {
    flex_free(b);
  }
  if(af -> back_ptr != a) return;
  af -> back_ptr = b;
  *a = 0;
  *b = &af -> data[0];
  _reg(FLEX_MOVE);
}

static int rmalloc_extend(int size)
{
  if(flex_extendby(& rmalloc_block, size) == 0)
    return 0;
  return size;
}

static void flex_clearblock(flex_block * b)
{
  if(b -> next) flex_clearblock(b -> next);
  if(b -> data != rmalloc_block) flex_free(b -> back_ptr);
}

void flex_clear(void)
{
  flex_clearblock(first);
  _reg(FLEX_CLEAR);
}

int flex_realloc(char * * b, int size)
{
  if(*b) {
    return flex_extend(b, size);
  } else {
    return flex_alloc(b, size);
  }
}

int flex_usingdynamic(void)
{
  return(dynamic_area != 0);
}

void flex_printchain(FILE * f)
{
  flex_block * current = first;
  int i = 0;

  if(f == 0) return;

  fprintf(f, "Flex:DynamicArea=%d\n", dynamic_area);
  fprintf(f, "Flex:first      =&%p\n", first);
  fprintf(f, "Flex:last       =&%p\n", last);
  fflush(f);
  fprintf(f, "## Location Anchor   Size     Next     Dec. size\n");
  while((current != last) && (current)) {
    fprintf(f, "%-.2d %p %p %-.8x %p %d\n", i, & current -> data[1], current -> back_ptr, current -> size, current -> next, current -> size);
    fflush(f);
    i ++ ;
    current = current -> next;
  }
  fprintf(f, "Flex:End of chain\n");
}
